/**************************************************************************
*  This file is part of the TCTS project (Tiny Cooperative Task Scheduler)
*
*  Copyright (c) 2014-2018 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  17.01.2014  mifi  First Version.
*  04.01.2016  mifi  Added more timezone functionality.
*  10.01.2016  mifi  Added OSTaskGetPriority and OSTimeDlyUntil.
*                    Added event functionality.
*  04.02.2016  mifi  Remove OSSemaBroadcast and OSSemaBroadcastAsync.
*                    Change OSSemaCreate parameter from uint32 to int32.
*  12.02.2016  mifi  Added the some more functionality:
*                    - OSTaskTerminateRequest and OSTaskShouldTerminate
*                    - OSTimerSet 
*                    - OSSemaReset
*  21.03.2016  mifi  Added OSSemaBroadcast and OSSemaBroadcastAsync again.
*  05.05.2016  mifi  Added incomplete Mailbox functionality, only:
*                    - OSMboxCreate
*                    - OSMboxDelete
*                    - OSMboxPostFromInt
*                    - OSMboxWait
*  16.06.2016  mifi  Added Cortex-M0+ support.
*  10.03.2018  mifi  Change version to v0.16.
*  12.05.2018  mifi  Added TaskScheduleExit support.
*                    Change version to v0.17.
*  16.05.2018  mifi  Check 8 byte stack alignment.
*                    Change version to v0.18.
*  23.06.2018  mifi  Change version pattern.
*                    Rework OS_WAIT_INFINITE handling.
*                    Added mutex functionality
*                    - OSMutexCreate
*                    - OSMutexSignal
*                    - OSMutexSignalFromInt
*                    - OSMutexWait
*                    Change version to v0.18.1
**************************************************************************/
#if !defined(__TCTS_H__)
#define __TCTS_H__

/**************************************************************************
*  Includes
**************************************************************************/
#include <stdint.h>
#include "tcts_conf.h"

/**************************************************************************
*  All Structures and Common Constants
**************************************************************************/

/*
 * Default ticks per second
 */
#if !defined(OS_TICKS_PER_SECOND)
#define OS_TICKS_PER_SECOND   1000
#endif 

/*
 * OS version information
 */
#define OS_VER_MAJOR          0
#define OS_VER_MINOR          18
#define OS_VER_PATCH          1

#define OS_VER_NUMBER         ((OS_VER_MAJOR << 24) | (OS_VER_MINOR << 16) | (OS_VER_PATCH << 8))

#define OS_VER_STRING         (XSTR(OS_VER_MAJOR) "." \
                               XSTR(OS_VER_MINOR) "." \
                               XSTR(OS_VER_PATCH))

#define OS_NAME               "TinyCTS" 


/*
 * Define for infinite timeout
 */
#define OS_WAIT_INFINITE   ((uint32_t)-1)


typedef void (*os_task_t)(void *arg);

typedef enum
{
   OS_TASK_STATE_NOT_IN_USE = 0,
   OS_TASK_STATE_CREATED,
   OS_TASK_STATE_READY,
   OS_TASK_STATE_RUNNING,
   OS_TASK_STATE_WAITING
} os_task_state_t;


/*
 * Forward references
 */
typedef struct _os_tcb_fifo_  os_tcb_fifo_t; 
typedef struct _os_tcb_       os_tcb_t;
typedef struct _os_sema_      os_sema_t; 
typedef struct _os_mutex_     os_mutex_t; 
typedef struct _os_event_     os_event_t; 


/*
 * Used to build a FIFO
 */
struct _os_tcb_fifo_
{
   os_tcb_t *pIn;
   os_tcb_t *pOut;
};


/*
 * Return code of OS functions
 */
#define OS_RC_OK       0
#define OS_RC_ERROR    -1
#define OS_RC_TIMEOUT  -2
#define OS_RC_NO_SPACE -3
#define OS_RC_DELETED  -4
#define OS_RC_RESETED  -5


/*
 * Semaphore
 */
#define OS_SEMA_COUNTER_MAX   INT32_MAX
struct _os_sema_
{
   os_tcb_fifo_t  Fifo;       /* Used for the tasks which are waiting for the semaphore */
   int32_t       nCounter;
   int32_t       nCounterMax;   
};


/*
 * Mutex
 */
#define OS_MUTEX_COUNTER_MAX   INT32_MAX
struct _os_mutex_
{
   os_tcb_fifo_t  Fifo;       /* Used for the tasks which are waiting for the mutex */
   os_tcb_t      *Owner;
   int32_t       nCounter;   
};


/*
 * Event
 */ 
typedef enum
{
   OS_EVENT_MODE_UNKNOWN = 0,
   OS_EVENT_MODE_OR,
   OS_EVENT_MODE_AND
} os_event_mode_t;
 
struct _os_event_
{
   os_tcb_fifo_t  Fifo;       /* Used for the tasks which are waiting for the event */
   uint32_t      dPattern;
}; 


/*
 * Mailbox
 */
typedef struct
{
   os_sema_t    UsedCntSema;
   os_sema_t    FreeCntSema;
   uint16_t    wCountMax;
   uint16_t    wCount;
   uint16_t    wInIndex;
   uint16_t    wOutIndex;
   void      **pBuffer;
} os_mbox_t;  


/*
 * Task Control Block
 */
struct _os_tcb_
{
   os_tcb_t        *pNext;                /* Used for the different lists */
   os_tcb_t        *pPrev;                /* Used for the different lists */

   os_tcb_t        *pWaitNext;            /* Used for the WaitList */
   os_tcb_t        *pWaitPrev;            /* Used for the WaitList */
   
   os_tcb_t        *pTaskNext;            /* Used for the TaskList */
   os_tcb_t        *pTaskPrev;            /* Used for the TaskList */
   
   char              Name[17];            /* Name */
   int              nPrio;                /* Priority */
   os_task_state_t   State;               /* State, for debug purpose */
   uintptr_t         StackPtr;            /* Actual stack pointer */
   uint8_t         *pStackStart;          /* Start of the stack */
   uint16_t         wStackSize;           /* Stack size */
   
   uint32_t         dTimeoutTicks;        /* Wait ticks in case of a timeout */
   
   os_sema_t       *pSemaWait;            /* Pointer to the semaphore if the task is waiting for it */
   
   os_mutex_t      *pMutexWait;           /* Pointer to the mutex if the task is waiting for it */
   
   os_event_t      *pEventWait;           /* Pointer to the event if the task is waiting for it */ 
   uint32_t         dEventWaitPattern;    /* Event pattern the task is waiting for */
   os_event_mode_t   EventWaitMode;       /* Event wait mode, OR or AND */
   
   int              nReturnCode;          /* Return value from OSSemaWait, OSEventWait */

   uint32_t         dStatStartTime;       /* Statistic start time */
   uint32_t         dStatEndTime;         /* Statistic end time */
   uint32_t         dStatTotalTime;       /* Statistic total task time */
   uint32_t         dStatTotalLastTime;   /* Statistic total task time of the previous statistic cycle */
   
   uint8_t          bFlagTermRequest;     /* Termination request flag */
};


typedef void (*OS_TIMER_FUNC)(void);

/**************************************************************************
*  Macro Definitions
**************************************************************************/

/*
 * This is the macro for defining the stack, use it like:
 *
 * static OS_TASK(MyStack, 512);
 *
 * The stack will be 8 byte aligned and the size 4 byte.
 */
#define OS_STACK(_n,_s)       uint8_t _n[((_s+3)& ~3)] __attribute__((aligned(0x8)))


/*
 * Some time management macros for converting MS => Ticks and Ticks => MS
 */
#if (OS_TICKS_PER_SECOND != 1000) 
#define OS_MS_2_TICKS(_a)     ((_a * OS_TICKS_PER_SECOND) / 1000)
#define OS_TICKS_2_MS(_a)     ((1000 / OS_TICKS_PER_SECOND) * _a)
#else
#define OS_MS_2_TICKS(_a)     (_a)
#define OS_TICKS_2_MS(_a)     (_a)
#endif


/*
 * Macro to check timeout
 *
 *    a = actual time
 *    s = start time
 *    t = timeout
 */
#define OS_TEST_TIMEOUT(_a,_s,_t)   (((uint32_t)(_a) - (uint32_t)(_s)) >= (uint32_t)(_t))


/*
 * Some resource macros
 */
#define OS_RES_CREATE(_a)     OSSemaCreate(_a, 1, 1)
#define OS_RES_LOCK(_a)       OSSemaWait(_a, OS_WAIT_INFINITE)
#define OS_RES_FREE(_a)       OSSemaSignal(_a)

/**************************************************************************
*  Functions Definitions
**************************************************************************/

/*
 * General functionality
 */
void      OSInit (void);
void      OSStart (void);
void      OSSysTickStart (void);

void      OSOutputRuntimeStackInfo (void);
void      OSOutputTaskInfo (void);
uint32_t  OSGetStackInfo (os_tcb_t *pTCB, uint32_t *pSize);


/*
 * Statistic functionality
 */
void      OSStatEnable (void);
uint8_t   OSStatGetCPULoad (void);
uint32_t  OSStatGetTotalTaskTime (void);


/*
 * Task functionality
 */
void      OSTaskCreate (os_tcb_t *pTCB, os_task_t Task, void *pParam, int nPrio,
                        uint8_t *pStack, uint16_t wStackSize, char *pName);
int       OSTaskGetPriority (void);
int       OSTaskChangePriority (int nPrio);
void      OSTaskYield (void);
void      OSTaskExit (void);
void      OSTaskTerminateRequest (os_tcb_t *pTCB);
uint8_t   OSTaskShouldTerminate (void);
os_tcb_t *OSTaskGetList (void);


/*
 * Time functionality
 */
void      OSTimerCallback (void);
void      OSTimerSet (OS_TIMER_FUNC Timer, uint32_t dTimeoutMs);

uint32_t  OSTimeGet (void);
uint32_t  OSTimeGetSeconds (void);
void      OSTimeDly (uint32_t dTimeoutMs);
void      OSTimeDlyUntil (uint32_t dTicks);
void      OSTimeWait (uint32_t dTimeoutMs);
uint32_t  OSUnixtimeGet (void);
void      OSUnixtimeSet (uint32_t dTime);
void      OSTimezoneIDSet (int16_t wID);
int16_t   OSTimezoneIDGet (void);
void      OSTimezoneMinSet (int16_t wOffset);
int16_t   OSTimezoneMinGet (void);
int32_t   OSTimezoneSecGet (void);
void      OSTimezoneDstSet (int8_t bOffset);
int32_t   OSTimezoneDstSecGet (void);


/*
 * Semaphore functionality
 */
void      OSSemaCreate (os_sema_t *pSema, int32_t nCounterStart, int32_t nCounterMax);
void      OSSemaReset (os_sema_t *pSema, int32_t nCounterStart);
void      OSSemaDelete (os_sema_t *pSema);
int       OSSemaSignal (os_sema_t *pSema);
int       OSSemaSignalFromInt (os_sema_t *pSema);
int       OSSemaSignalNoSched (os_sema_t *pSema);
int       OSSemaWait (os_sema_t *pSema, uint32_t dTimeoutMs);

/* Some functions needed by NutNET */
int       OSSemaBroadcast (os_sema_t *pSema);
int       OSSemaBroadcastAsync (os_sema_t *pSema);  


/*
 * Mutex functionality
 */
void      OSMutexCreate (os_mutex_t *pMutex);
int       OSMutexSignal (os_mutex_t *pMutex);
int       OSMutexSignalFromInt (os_mutex_t *pMutex);
int       OSMutexWait (os_mutex_t *pMutex, uint32_t dTimeoutMs);


/*
 * Event functionality
 */ 
void      OSEventCreate (os_event_t *pEvent);
void      OSEventDelete (os_event_t *pEvent);
void      OSEventSet (os_event_t *pEvent, uint32_t dPattern);
void      OSEventSetFromInt (os_event_t *pEvent, uint32_t dPattern);
void      OSEventSetNoSched (os_event_t *pEvent, uint32_t dPattern);
int       OSEventWait (os_event_t *pEvent, uint32_t dWaitPattern, 
                       os_event_mode_t Mode, uint32_t *pPattern, uint32_t dTimeoutMs);

                       
/*
 * Mailbox functionality
 */       
void      OSMboxCreate (os_mbox_t *pMbox, void **pBuffer, uint16_t wCounterMax);
void      OSMboxDelete (os_mbox_t *pMbox);
int       OSMboxPostFromInt (os_mbox_t *pMbox, void *pMsg);
int       OSMboxWait (os_mbox_t *pMbox, void **pMsg, uint32_t dTimeoutMs);

#endif /* !__TCTS_H__ */

/*** EOF ***/
